// Copyright 2021 Sony Corporation. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

#ifndef FLUTTER_SHELL_PLATFORM_LINUX_EMBEDDED_WINDOW_ELINUX_WINDOW_WAYLAND_H_
#define FLUTTER_SHELL_PLATFORM_LINUX_EMBEDDED_WINDOW_ELINUX_WINDOW_WAYLAND_H_

#include <wayland-client.h>
#include <wayland-cursor.h>

#include <memory>
#include <unordered_map>
#include <vector>

#include "flutter/shell/platform/linux_embedded/surface/surface_gl.h"
#include "flutter/shell/platform/linux_embedded/window/elinux_window.h"
#include "flutter/shell/platform/linux_embedded/window/native_window_wayland.h"
#include "flutter/shell/platform/linux_embedded/window/renderer/window_decorations_wayland.h"
#include "flutter/shell/platform/linux_embedded/window_binding_handler.h"

// These header files are automatically generated by the
// wayland-scanner.
extern "C" {
#include "wayland/protocols/presentation-time-protocol.h"
#include "wayland/protocols/text-input-unstable-v1-client-protocol.h"
#include "wayland/protocols/text-input-unstable-v3-client-protocol.h"
#include "wayland/protocols/xdg-shell-client-protocol.h"
}

namespace flutter {

namespace {
constexpr char kXcursorSizeEnvironmentKey[] = "XCURSOR_SIZE";
}  // namespace

class ELinuxWindowWayland : public ELinuxWindow, public WindowBindingHandler {
 public:
  ELinuxWindowWayland(FlutterDesktopViewProperties view_properties);
  ~ELinuxWindowWayland();

  // |ELinuxWindow|
  bool IsValid() const override;

  // |FlutterWindowBindingHandler|
  bool DispatchEvent() override;

  // |FlutterWindowBindingHandler|
  bool CreateRenderSurface(int32_t width, int32_t height) override;

  // |FlutterWindowBindingHandler|
  void DestroyRenderSurface() override;

  // |FlutterWindowBindingHandler|
  void SetView(WindowBindingHandlerDelegate* view) override;

  // |FlutterWindowBindingHandler|
  ELinuxRenderSurfaceTarget* GetRenderSurfaceTarget() const override;

  // |FlutterWindowBindingHandler|
  uint16_t GetRotationDegree() const override;

  // |FlutterWindowBindingHandler|
  double GetDpiScale() override;

  // |FlutterWindowBindingHandler|
  PhysicalWindowBounds GetPhysicalWindowBounds() override;

  // |FlutterWindowBindingHandler|
  int32_t GetFrameRate() override;

  // |FlutterWindowBindingHandler|
  void UpdateFlutterCursor(const std::string& cursor_name) override;

  // |FlutterWindowBindingHandler|
  void UpdateVirtualKeyboardStatus(const bool show) override;

  // |FlutterWindowBindingHandler|
  std::string GetClipboardData() override;

  // |FlutterWindowBindingHandler|
  void SetClipboardData(const std::string& data) override;

 private:
  struct CursorInfo {
    std::string cursor_name;
    uint32_t serial;
    wl_pointer* pointer;
  };

  void WlRegistryHandler(wl_registry* wl_registry,
                         uint32_t name,
                         const char* interface,
                         uint32_t version);

  void WlUnRegistryHandler(wl_registry* wl_registry, uint32_t name);

  bool LoadCursorTheme(uint32_t size);

  wl_cursor* GetWlCursor(const std::string& cursor_name, uint32_t size);

  void ShowVirtualKeyboard();

  void DismissVirtualKeybaord();

  static const wl_registry_listener kWlRegistryListener;
  static const xdg_wm_base_listener kXdgWmBaseListener;
  static const xdg_surface_listener kXdgSurfaceListener;
  static const xdg_toplevel_listener kXdgToplevelListener;
  static const wl_seat_listener kWlSeatListener;
  static const wl_pointer_listener kWlPointerListener;
  static const wl_touch_listener kWlTouchListener;
  static const wl_keyboard_listener kWlKeyboardListener;
  static const wl_output_listener kWlOutputListener;
  static const wl_data_device_listener kWlDataDeviceListener;
  static const wl_data_source_listener kWlDataSourceListener;
  static const zwp_text_input_v1_listener kZwpTextInputV1Listener;
  static const zwp_text_input_v3_listener kZwpTextInputV3Listener;
  static const wl_callback_listener kWlSurfaceFrameListener;
  static const wp_presentation_listener kWpPresentationListener;
  static const wp_presentation_feedback_listener
      kWpPresentationFeedbackListener;
  static constexpr size_t kDefaultPointerSize = 24;

  // A pointer to a FlutterWindowsView that can be used to update engine
  // windowing and input state.
  WindowBindingHandlerDelegate* binding_handler_delegate_ = nullptr;

  std::unique_ptr<NativeWindowWayland> native_window_;
  std::unique_ptr<SurfaceGl> render_surface_;

  // decorations.
  std::unique_ptr<WindowDecorationsWayland> window_decorations_;
  wl_surface* wl_current_surface_;
  wl_subcompositor* wl_subcompositor_;
  bool restore_window_required_ = false;
  int32_t restore_window_width_;
  int32_t restore_window_height_;

  bool display_valid_;
  bool running_;
  bool maximised_;
  uint32_t last_frame_time_;

  // Indicates that exists a keyboard show request from Flutter Engine.
  bool is_requested_show_virtual_keyboard_;

  wl_display* wl_display_;
  wl_registry* wl_registry_;
  wl_compositor* wl_compositor_;
  wl_seat* wl_seat_;
  wl_output* wl_output_;
  wl_shm* wl_shm_;
  wl_pointer* wl_pointer_;
  wl_touch* wl_touch_;
  wl_keyboard* wl_keyboard_;
  wl_surface* wl_cursor_surface_;
  xdg_wm_base* xdg_wm_base_;
  xdg_surface* xdg_surface_;
  xdg_toplevel* xdg_toplevel_;

  // text-input protocol for onscreen keyboard inputs.
  zwp_text_input_manager_v1* zwp_text_input_manager_v1_;
  zwp_text_input_manager_v3* zwp_text_input_manager_v3_;
  zwp_text_input_v1* zwp_text_input_v1_;
  zwp_text_input_v3* zwp_text_input_v3_;

  // Frame information for Vsync events.
  wp_presentation* wp_presentation_;
  uint32_t wp_presentation_clk_id_;
  uint64_t last_frame_time_nanos_;
  int32_t frame_rate_;

  CursorInfo cursor_info_;
  size_t cursor_size_;

  // List of cursor name and wl_cursor supported by Wayland keyed by their
  // (scaled) size.
  std::unordered_map<uint32_t, std::unordered_map<std::string, wl_cursor*>>
      supported_wl_cursor_list_;

  // List of loaded Wayland cursor themes keyed by their (scaled) size.
  std::unordered_map<uint32_t, wl_cursor_theme*> wl_cursor_themes_;

  wl_data_device_manager* wl_data_device_manager_;
  wl_data_device* wl_data_device_;
  wl_data_offer* wl_data_offer_;
  wl_data_source* wl_data_source_;
  uint32_t wl_data_device_manager_version_;
  uint32_t serial_;
};

}  // namespace flutter

#endif  // FLUTTER_SHELL_PLATFORM_LINUX_EMBEDDED_WINDOW_ELINUX_WINDOW_WAYLAND_H_
